@@ -1,6 +1,6 @@
 # Orca
 #
-# Copyright 2005-2007 Sun Microsystems Inc.
+# Copyright 2005-2008 Sun Microsystems Inc.
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Library General Public
@@ -14,8 +14,8 @@
 #
 # You should have received a copy of the GNU Library General Public
 # License along with this library; if not, write to the
-# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-# Boston, MA 02111-1307, USA.
+# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
+# Boston MA  02110-1301 USA.
 
 """A very experimental approach to the refreshable Braille display.  This
 module treats each line of the display as a sequential set of regions, where
@@ -25,10 +25,10 @@ operations on the Accessible object, such as invoking default actions or
 moving the text caret.
 """
 
-__id__        = "$Id: braille.py 3571 2008-02-12 23:27:57Z eitani $"
-__version__   = "$Revision: 3571 $"
-__date__      = "$Date: 2008-02-12 18:27:57 -0500 (Tue, 12 Feb 2008) $"
-__copyright__ = "Copyright (c) 2005-2007 Sun Microsystems Inc."
+__id__        = "$Id: braille.py 4034 2008-07-14 19:09:34Z wwalker $"
+__version__   = "$Revision: 4034 $"
+__date__      = "$Date: 2008-07-14 15:09:34 -0400 (Mon, 14 Jul 2008) $"
+__copyright__ = "Copyright (c) 2005-2008 Sun Microsystems Inc."
 __license__   = "LGPL"
 
 import logging
@@ -36,8 +36,6 @@ log = logging.getLogger("braille")
 
 import signal
 
-import pyatspi
-
 try:
     import louis
 except ImportError:
@@ -454,7 +452,8 @@ class Text(Region):
     [[[TODO: WDW - need to add in text selection capabilities.  Logged
     as bugzilla bug 319754.]]]"""
 
-    def __init__(self, accessible, label="", eol=""):
+    def __init__(self, accessible, label="", eol="", 
+                 startOffset=None, endOffset=None):
         """Creates a new Text region.
 
         Arguments:
@@ -466,20 +465,24 @@ class Text(Region):
         if orca_state.activeScript:
             [string, self.caretOffset, self.lineOffset] = \
                  orca_state.activeScript.getTextLineAtCaret(self.accessible)
+        else:
+            string = ""
 
-        # Sometimes, gnome-terminal will give us very odd values when
-        # the user is editing using 'vi' and has positioned the caret
-        # at the first character of the first line.  In this case, we
-        # end up getting a very large negative number for the line offset.
-        # So, we just assume the user is at the first character.
-        #
-        if self.lineOffset < 0:
-            self.caretOffset = 0
-            self.lineOffset = 0
-            [string, startOffset, endOffset] = \
-                self.accessible.queryText().getTextAtOffset(
-                    0,
-                    pyatspi.TEXT_BOUNDARY_LINE_START)
+        string = string.decode("UTF-8")
+
+        try:
+            endOffset = endOffset - self.lineOffset
+        except TypeError:
+            pass
+
+        try:
+            self.startOffset = startOffset - self.lineOffset
+        except TypeError:
+            self.startOffset = 0
+
+        string = string[self.startOffset:endOffset]
+
+        self.caretOffset -= self.startOffset
 
         cursorOffset = min(self.caretOffset - self.lineOffset, len(string))
 
@@ -495,10 +498,10 @@ class Text(Region):
         string = self.label + string
 
         cursorOffset += len(self.label)
-        
+
         Region.__init__(self, string, cursorOffset, True)
 
-        if not self.contracted:
+        if not self.contracted and not settings.disableBrailleEOL:
             self.string += self.eol
 
     def repositionCursor(self):
@@ -540,7 +543,8 @@ class Text(Region):
             return
 
         newCaretOffset = min(self.lineOffset + offset, self._maxCaretOffset)
-        self.accessible.queryText().setCaretOffset(newCaretOffset)
+        orca_state.activeScript.setCaretOffset(
+            self.accessible, newCaretOffset)
 
     def getAttributeMask(self):
         """Creates a string which can be used as the attrOr field of brltty's
@@ -554,84 +558,64 @@ class Text(Region):
 
         # Start with an empty mask.
         #
-        stringLength = len(self.string) - len(self.eol)
-        attrMask = ['\x00']*stringLength
+        stringLength = len(self.rawLine) - len(self.label)
+        lineEndOffset = self.lineOffset + stringLength
+        regionMask = [settings.TEXT_ATTR_BRAILLE_NONE]*stringLength
 
         attrIndicator = settings.textAttributesBrailleIndicator
         selIndicator = settings.brailleSelectorIndicator
-        indicateAttr = (attrIndicator != settings.TEXT_ATTR_BRAILLE_NONE)
-        indicateSel = (selIndicator != settings.BRAILLE_SEL_NONE)
+        linkIndicator = settings.brailleLinkIndicator
+        script = orca_state.activeScript
 
-        if indicateAttr and settings.enabledBrailledTextAttributes:
-            # Identify what attributes the user cares about.  Also get the
-            # default attributes for the text object because those attributes
-            # may not be included as attributes for the character.
-            #
-            script = orca_state.activeScript
-            enabledAttributes = settings.enabledBrailledTextAttributes
-            [userAttrList, userAttrDict] = \
-                script.textAttrsToDictionary(enabledAttributes)
-            defaultAttributes = text.getDefaultAttributes()
-            [defaultAttrList, defaultAttrDict] = \
-                script.textAttrsToDictionary(defaultAttributes)
+        if linkIndicator != settings.BRAILLE_LINK_NONE:
+            try:
+                hyperText = self.accessible.queryHypertext()
+                nLinks = hyperText.getNLinks()
+            except:
+                nLinks = 0
 
-            # [[[TODO - HACK - JD: The end offset returned by SO/OOo is the
-            # offset of the first character that doesn't have the attribute
-            # in question; all other apps return the offset of the last
-            # character that has that attribute]]]
-            #
-            adjustment = 0
-            appName = orca_state.locusOfFocus.getApplication().name
-            if appName.startswith("soffice"):
-                adjustment += 1
+            n = 0
+            while n < nLinks:
+                link = hyperText.getLink(n)
+                if self.lineOffset <= link.startIndex:
+                    for i in xrange(link.startIndex, link.endIndex):
+                        try:
+                            regionMask[i] |= linkIndicator
+                        except:
+                            pass
+                n += 1
 
-            i = self.lineOffset
-            endOffset = self.lineOffset + stringLength
-            while i < endOffset - 1:
-                for attribute in userAttrList:
-                    defaultValue = None
-                    if attribute in defaultAttrList:
-                        defaultValue = defaultAttrDict[attribute]
-                    [value, start, end, defined] = \
-                            text.getAttributeValue(i, attribute)
+        if attrIndicator:
+            enabledAttributes = script.attribsToDictionary(
+                settings.enabledBrailledTextAttributes)
 
-                    notOfInterest = userAttrDict[attribute]
-                    if defined:
-                        weCare = (value != notOfInterest)
-                    elif defaultValue and len(notOfInterest):
-                        weCare = (defaultValue != notOfInterest)
-                    else:
-                        weCare = False
-
-                    if weCare and start >= 0:
-                        maskStart = start - self.lineOffset
-                        maskEnd = end - self.lineOffset - adjustment
-                        maskLength = maskEnd - maskStart
-                        attrMask[maskStart:maskEnd] = \
-                                                  [attrIndicator] * maskLength
-                if (end <= i):
+            offset = self.lineOffset
+            while offset < lineEndOffset:
+                attributes, startOffset, endOffset = \
+                            script.getTextAttributes(self.accessible,
+                                                     offset, True)
+                if endOffset <= offset:
                     break
+                mask = settings.TEXT_ATTR_BRAILLE_NONE
+                offset = endOffset
+                for attrib in attributes:
+                    if enabledAttributes.get(attrib, '') != '':
+                        if enabledAttributes[attrib] != attributes[attrib]:
+                            mask = attrIndicator
+                            break
+                if mask != settings.TEXT_ATTR_BRAILLE_NONE:
+                    maskStart = max(startOffset - self.lineOffset, 0)
+                    maskEnd = min(endOffset - self.lineOffset, stringLength)
+                    for i in xrange(maskStart, maskEnd):
+                        regionMask[i] |= attrIndicator
 
-                i = end
-
-        if indicateSel:
-            nSelections = text.getNSelections()
-            if nSelections:
-                for i in range(0, nSelections):
-                    [start, end] = text.getSelection(i)
-                    maskStart = max(self.lineOffset, start) - self.lineOffset
-                    maskEnd = min(end - self.lineOffset, stringLength)
-                    # Combine the selection indicator with the attribute
-                    # indicator.
-                    #
-                    for j in range(maskStart, maskEnd):
-                        if attrMask[j] != '\x00' \
-                           and attrMask[j] != selIndicator:
-                            attrMask[j] = '\xc0'
-                        else:
-                            attrMask[j] = selIndicator
-
-        mask = "".join(attrMask)
+        if selIndicator:
+            selections = script.getTextSelections(self.accessible)
+            for startOffset, endOffset in selections:
+                maskStart = max(startOffset - self.lineOffset, 0)
+                maskEnd = min(endOffset - self.lineOffset, stringLength)
+                for i in xrange(maskStart, maskEnd):
+                    regionMask[i] |= selIndicator
 
         if self.contracted:
             contractedMask = [0] * len(self.rawLine)
@@ -640,24 +624,22 @@ class Text(Region):
                 # Transform the offsets.
                 outPos = \
                        [offset - len(self.label) - 1 for offset in outPos]
-            for i, m in enumerate(mask):
+            for i, m in enumerate(regionMask):
                 try:
-                    contractedMask[outPos[i]] |= ord(m)
+                    contractedMask[outPos[i]] |= m
                 except IndexError:
                     continue
-            mask = ''.join(map(chr, contractedMask))
+            regionMask = contractedMask
 
         # Add empty mask characters for the EOL character as well as for
         # any label that might be present.
         #
-        for i in range (0, len(self.eol)):
-            mask += '\x00'
+        regionMask += [0]*len(self.eol)
 
         if self.label:
-            for i in range (0, len(self.label) + 1):
-                mask = '\x00' + mask
+            regionMask = [0]*len(self.label) + regionMask
 
-        return mask
+        return ''.join(map(chr, regionMask))
 
     def contractLine(self, line, cursorOffset=0, expandOnCursor=True):
         contracted, inPos, outPos, cursorPos = Region.contractLine(
@@ -667,6 +649,7 @@ class Text(Region):
 
     def displayToBufferOffset(self, display_offset):
         offset = Region.displayToBufferOffset(self, display_offset)
+        offset += self.startOffset
         offset -= len(self.label)
         return offset
 
@@ -688,7 +671,6 @@ class ReviewComponent(Component):
                         for this Region if it gets focus.
         - zone: the flat review Zone associated with this component
         """
-
         Component.__init__(self, accessible, string,
                            cursorOffset, expandOnCursor=True)
         self.zone = zone
@@ -709,7 +691,6 @@ class ReviewText(Region):
         - lineOffset: the character offset into where the text line starts
         - zone: the flat review Zone associated with this component
         """
-
         Region.__init__(self, string, expandOnCursor=True)
         self.accessible = accessible
         self.lineOffset = lineOffset
@@ -723,7 +704,7 @@ class ReviewText(Region):
 
         offset = self.displayToBufferOffset(offset)
         newCaretOffset = self.lineOffset + offset
-        self.accessible.queryText().setCaretOffset(newCaretOffset)
+        orca_state.activeScript.setCaretOffset(self.accessible, newCaretOffset)
 
 class Line:
     """A horizontal line on the display.  Each Line is composed of a sequential
